/*
	PIC Metal Detector - Frequency Counting PIC Based Metal Detector
	Copyright (C) 2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

#define __16f648a
#include <pic16f648a.h>

#include "counter.h"
#include "eeprom.h"
#include "led.h"
#include "pwm.h"
#include "serial.h"

/* Configuration */
unsigned int __at 0x2007 __CONFIG =
	_CP_OFF &
	_CPD_OFF & 
	_BODEN_OFF &
	_LVP_OFF &
	_MCLRE_OFF &
	_PWRTE_OFF &
	_WDT_OFF &
	_INTRC_OSC_CLKOUT;

/* State */
static unsigned char cont_readout = 0;
static unsigned char count_pwm = 0;

static unsigned char refadj = 0;
static int reference = 0;

/* Functions */
static void DoCount()
{
	unsigned int count = Count(eeprom_read_int(EEPROM_COUNT_WAIT));
	serial_transmit_string("0x");
	serial_transmit_byte(count >> 8);
	serial_transmit_byte(count);
	serial_transmit_string("\n");
}

static void DoCountPwm()
{
	int count = Count(eeprom_read_int(EEPROM_COUNT_WAIT));
	int diff = count - reference;
	
	/* Calculate the difference */
	if (diff < -128){
		diff = -128;
	}
	if (diff > 127){
		diff = 127;
	}
	diff = 127 - diff;
	
	
	/* Set the PWM frequency */
	pwm_init(diff, diff << 1, 1);
	
	/* Update the reference */
	if (!reference){ //initialise
		reference = count;
	}
	else {
		refadj++;
		if (eeprom_read_int(EEPROM_REFERENCE_ADJUST) == refadj){
			if (reference < count){
				reference++;
			}
			else if (reference > count){
				reference--;
			}
			refadj = 0;
		}
	}
}

static void process_pwm_start_stop()
{
	serial_transmit_string(count_pwm ? "Disabling PWM\n" : "Enabling PWM\n");
	reference = 0;
	count_pwm = !count_pwm;
	eeprom_write_byte(EEPROM_COUNT_PWM_ON_STARTUP, count_pwm);
}

static void cmd_eeprom_load()
{
	char c = 0;
	serial_transmit_string("Load from EEPROM: ");
	c = serial_receive_char(0);
	serial_transmit_string("\n");
	
	switch (c)
	{
		case 'c':
			serial_transmit_string("Count wait time: 0x");
			serial_transmit_int(eeprom_read_int(EEPROM_COUNT_WAIT));
			serial_transmit_string("\n");
			break;
		case 'r':
			serial_transmit_string("Reference adjust: 0x");
			serial_transmit_int(eeprom_read_int(EEPROM_REFERENCE_ADJUST));
			serial_transmit_string("\n");
			break;
		default:
			serial_transmit_string("Unknown EEPROM data: ");
			serial_transmit_char(c);
			serial_transmit_string("\n");
	}
}

static void cmd_eeprom_store()
{
	char c = 0;
	serial_transmit_string("Store to EEPROM: ");
	c = serial_receive_char(0);
	serial_transmit_string("\n");
	
	switch (c)
	{
		case 'c':
			serial_transmit_string("Count wait time: 0x");
			eeprom_write_int(EEPROM_COUNT_WAIT, serial_receive_int());
			serial_transmit_string("\n");
			break;
		case 'r':
			serial_transmit_string("Reference adjust: 0x");
			eeprom_write_int(EEPROM_REFERENCE_ADJUST, serial_receive_int());
			serial_transmit_string("\n");
			break;
		default:
			serial_transmit_string("Unknown EEPROM data: ");
			serial_transmit_char(c);
			serial_transmit_string("\n");
	}
}

static void process_cmd()
{
	unsigned char period = 0; //also pulse width
	unsigned char duty = 0; //also gap width
	unsigned char max_duty = 0;
	
	char c = serial_receive_char(0);
	switch (c)
	{
		case 'C':
			cont_readout = !cont_readout;
			break;
		case 'c':
			led_on(1);
			DoCount();
			led_off(1);
			break;
		case 'L': //Load information from EEPROM
			cmd_eeprom_load();
			break;
		case 'p':
			process_pwm_start_stop();
			break;
		case 'S': //Store information to EEPROM
			cmd_eeprom_store();
			break;
		case 'T':
			serial_transmit_string("Timestamp: ");
			serial_transmit_string(TIMESTAMP);
			serial_transmit_string("\n");
			break;
		default:
			serial_transmit_string("Unknown command: ");
			serial_transmit_char(c);
			serial_transmit_string("\n");
	}
}

/* Interrupt and Reset Functions */
void main()
{
	init_led();
	led_on(0);
	
	serial_init(); //Initialise USART
	serial_transmit_string("Hello, RS-232 TTL!\n");
	
	INTCON = _GIE | _PEIE; //Enable interrupts
	INTCON |= _INTE; //Enable PORTB[0] interrupt
	
	count_pwm = eeprom_read_byte(EEPROM_COUNT_PWM_ON_STARTUP);
	
	while (1){
		if (cont_readout){
			DoCount();
		}
		
		if (count_pwm){
			led_on(2);
			//serial_transmit_string("01\n");
			DoCountPwm();
			//serial_transmit_string("02\n");
		}
		else {
			pwm_deinit();
			led_off(2);
		}
	}
}

static void isr() __interrupt 0
{
	if (INTCON & _INTF){
		process_pwm_start_stop();
	}
	
	if (PIR1 & _RCIF){
		process_cmd();
	}
	
	INTCON &= 0xF8;
	PIR1 = 0;
}



